# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cvcuda
import pytest as t
import numpy as np
import util


RNG = np.random.default_rng(0)


@t.mark.parametrize(
    "input,base,scale,globalscale,globalshift,epsilon,flags",
    [
        (
            cvcuda.Tensor([5, 16, 23, 4], np.uint8, "NHWC"),
            cvcuda.Tensor([1, 1], np.float32, "HW"),
            cvcuda.Tensor([1, 1], np.float32, "HW"),
            1,
            2,
            3,
            None,
        ),
        (
            cvcuda.Tensor([5, 16, 23, 4], np.uint8, "NHWC"),
            cvcuda.Tensor([16, 1], np.float32, "HW"),
            cvcuda.Tensor([16, 1], np.float32, "HW"),
            1,
            2,
            3,
            cvcuda.NormalizeFlags.SCALE_IS_STDDEV,
        ),
        (
            cvcuda.Tensor([5, 16, 23, 4], np.uint8, "NHWC"),
            cvcuda.Tensor([1, 23], np.float32, "HW"),
            cvcuda.Tensor([1, 23], np.float32, "HW"),
            1,
            2,
            3,
            None,
        ),
        (
            cvcuda.Tensor([5, 16, 23, 4], np.uint8, "NHWC"),
            cvcuda.Tensor([16, 23], np.float32, "HW"),
            cvcuda.Tensor([16, 23], np.float32, "HW"),
            1,
            2,
            3,
            None,
        ),
    ],
)
def test_op_normalize(input, base, scale, globalscale, globalshift, epsilon, flags):
    out = cvcuda.normalize(input, base, scale)
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == input.dtype

    out = cvcuda.Tensor(input.shape, input.dtype, input.layout)
    tmp = cvcuda.normalize_into(out, input, base, scale)
    assert tmp is out
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == input.dtype

    stream = cvcuda.Stream()
    out = cvcuda.normalize(
        src=input,
        base=base,
        scale=scale,
        flags=flags,
        globalscale=globalscale,
        globalshift=globalshift,
        epsilon=epsilon,
        stream=stream,
    )
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == input.dtype

    tmp = cvcuda.normalize_into(
        src=input,
        dst=out,
        base=base,
        scale=scale,
        flags=flags,
        globalscale=globalscale,
        globalshift=globalshift,
        epsilon=epsilon,
        stream=stream,
    )
    assert tmp is out
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == input.dtype


@t.mark.parametrize(
    "nimages,format,max_size,max_pixel,base,scale,globalscale,globalshift,epsilon,flags",
    [
        (
            5,
            cvcuda.Format.RGB8,
            (16, 23),
            128.0,
            cvcuda.Tensor([1, 1, 1, 5], np.float32, "NHWC"),
            cvcuda.Tensor([1, 1, 1, 5], np.float32, "NHWC"),
            1,
            2,
            3,
            None,
        ),
        (
            5,
            cvcuda.Format.RGB8,
            (16, 23),
            256.0,
            cvcuda.Tensor([1, 1, 1, 5], np.float32, "NHWC"),
            cvcuda.Tensor([1, 1, 1, 5], np.float32, "NHWC"),
            1,
            2,
            3,
            cvcuda.NormalizeFlags.SCALE_IS_STDDEV,
        ),
    ],
)
def test_op_rotatevarshape(
    nimages,
    format,
    max_size,
    max_pixel,
    base,
    scale,
    globalscale,
    globalshift,
    epsilon,
    flags,
):
    input = util.create_image_batch(
        nimages, format, max_size=max_size, max_random=max_pixel, rng=RNG
    )

    out = cvcuda.normalize(input, base, scale)
    assert len(out) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == input.uniqueformat
    assert out.maxsize == input.maxsize

    out = util.clone_image_batch(input)
    tmp = cvcuda.normalize_into(out, input, base, scale)
    assert tmp is out
    assert len(out) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == input.uniqueformat
    assert out.maxsize == input.maxsize

    stream = cvcuda.cuda.Stream()
    out = cvcuda.normalize(
        src=input,
        base=base,
        scale=scale,
        flags=flags,
        globalscale=globalscale,
        globalshift=globalshift,
        epsilon=epsilon,
        stream=stream,
    )
    assert len(out) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == input.uniqueformat
    assert out.maxsize == input.maxsize

    tmp = cvcuda.normalize_into(
        src=input,
        dst=out,
        base=base,
        scale=scale,
        flags=flags,
        globalscale=globalscale,
        globalshift=globalshift,
        epsilon=epsilon,
        stream=stream,
    )
    assert tmp is out
    assert len(out) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == input.uniqueformat
    assert out.maxsize == input.maxsize



import os
import sys

# Internal build script for generating debugging test .so size.
# Usage:
#     python libsize.py file.so save.txt -- displays the size of file.so and, if save.txt exists, compares it to the
#                                           size in it, then overwrites save.txt with the new size for future runs.

if len(sys.argv) != 3:
    sys.exit("Invalid arguments: usage: python libsize.py file.so save.txt")

lib = sys.argv[1]
save = sys.argv[2]

if not os.path.exists(lib):
    sys.exit(f"Error: requested file ({lib}) does not exist")

libsize = os.path.getsize(lib)

print("------", os.path.basename(lib), "file size:", libsize, end="")

if os.path.exists(save):
    with open(save) as sf:
        oldsize = int(sf.readline())

    if oldsize > 0:
        change = libsize - oldsize
        if change == 0:
            print(" (no change)")
        else:
            print(f" (change of {change:+} bytes = {change / oldsize:+.2%})")
else:
    print()

with open(save, "w") as sf:
    sf.write(str(libsize))



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from torchvision import models

# Get pretrained model
resnet50 = models.resnet50(pretrained=True)

# Export the model to ONNX
inputWidth = 224
inputHeight = 224
maxBatchSize = 32
x = torch.randn(maxBatchSize, 3, inputHeight, inputWidth, requires_grad=True)
torch.onnx.export(
    resnet50,
    x,
    "./models/resnet50.onnx",
    export_params=True,
    opset_version=12,
    do_constant_folding=True,
    input_names=["input"],
    output_names=["output"],
    dynamic_axes={"input": {0: "maxBatchSize"}, "output": {0: "maxBatchSize"}},
)



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cvcuda
import pytest as t
import numpy as np


@t.mark.parametrize(
    "input,dtype,scale,offset",
    [
        (cvcuda.Tensor([5, 16, 23, 4], np.uint8, "NHWC"), np.float32, 1.2, 10.2),
        (cvcuda.Tensor([16, 23, 2], np.uint8, "HWC"), np.int32, -1.2, -5.5),
    ],
)
def test_op_convertto(input, dtype, scale, offset):
    out = cvcuda.convertto(input, dtype)
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == dtype

    out = cvcuda.Tensor(input.shape, dtype, input.layout)
    tmp = cvcuda.convertto_into(out, input)
    assert tmp is out
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == dtype

    out = cvcuda.convertto(input, dtype, scale)
    out = cvcuda.convertto(input, dtype, scale, offset)

    out = cvcuda.Tensor(input.shape, dtype, input.layout)
    tmp = cvcuda.convertto_into(out, input, scale)
    tmp = cvcuda.convertto_into(out, input, scale, offset)

    stream = cvcuda.Stream()
    out = cvcuda.convertto(
        src=input, dtype=dtype, scale=scale, offset=offset, stream=stream
    )
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == dtype

    tmp = cvcuda.convertto_into(
        dst=out, src=input, scale=scale, offset=offset, stream=stream
    )
    assert tmp is out
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == dtype



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cvcuda
import pytest as t
import numpy as np
import util


RNG = np.random.default_rng(0)


@t.mark.parametrize(
    "input, code, output",
    [
        (
            cvcuda.Tensor(5, [16, 23], cvcuda.Format.BGR8),
            cvcuda.ColorConversion.BGR2RGB,
            cvcuda.Tensor(5, [16, 23], cvcuda.Format.RGB8),
        ),
        (
            cvcuda.Tensor(3, [86, 22], cvcuda.Format.RGBA8),
            cvcuda.ColorConversion.RGBA2BGRA,
            cvcuda.Tensor(3, [86, 22], cvcuda.Format.BGRA8),
        ),
        (
            cvcuda.Tensor(7, [13, 21], cvcuda.Format.Y8),
            cvcuda.ColorConversion.GRAY2BGR,
            cvcuda.Tensor(7, [13, 21], cvcuda.Format.BGR8),
        ),
        (
            cvcuda.Tensor(9, [66, 99], cvcuda.Format.HSV8),
            cvcuda.ColorConversion.HSV2RGB,
            cvcuda.Tensor(9, [66, 99], cvcuda.Format.RGB8),
        ),
        (
            cvcuda.Tensor([1, 61, 62, 3], np.uint8, "NHWC"),
            cvcuda.ColorConversion.YUV2RGB,
            cvcuda.Tensor([1, 61, 62, 3], np.uint8, "NHWC"),
        ),
    ],
)
def test_op_cvtcolor(input, code, output):
    out = cvcuda.cvtcolor(input, code)
    assert out.shape == output.shape
    assert out.dtype == output.dtype

    stream = cvcuda.Stream()
    tmp = cvcuda.cvtcolor_into(
        src=input,
        dst=output,
        code=code,
        stream=stream,
    )
    assert tmp is output
    assert output.shape[:-1] == input.shape[:-1]


@t.mark.parametrize(
    "num_images, in_format, img_size, max_pixel, code, out_format",
    [
        (
            10,
            cvcuda.Format.RGB8,
            (123, 321),
            256,
            cvcuda.ColorConversion.RGB2RGBA,
            cvcuda.Format.RGBA8,
        ),
        (
            8,
            cvcuda.Format.BGRA8,
            (23, 21),
            256,
            cvcuda.ColorConversion.BGRA2RGB,
            cvcuda.Format.RGB8,
        ),
        (
            6,
            cvcuda.Format.RGB8,
            (23, 21),
            256,
            cvcuda.ColorConversion.RGB2GRAY,
            cvcuda.Format.Y8_ER,
        ),
        (
            4,
            cvcuda.Format.HSV8,
            (23, 21),
            256,
            cvcuda.ColorConversion.HSV2RGB,
            cvcuda.Format.RGB8,
        ),
        (
            2,
            cvcuda.Format.Y8_ER,
            (23, 21),
            256,
            cvcuda.ColorConversion.GRAY2BGR,
            cvcuda.Format.BGR8,
        ),
    ],
)
def test_op_cvtcolorvarshape(
    num_images, in_format, img_size, max_pixel, code, out_format
):
    input = util.create_image_batch(
        num_images, in_format, size=img_size, max_random=max_pixel, rng=RNG
    )
    output = util.create_image_batch(
        num_images, out_format, size=img_size, max_random=max_pixel, rng=RNG
    )
    out = cvcuda.cvtcolor(input, code)
    assert len(out) == len(output)
    assert out.capacity == output.capacity
    assert out.maxsize == output.maxsize

    stream = cvcuda.Stream()
    tmp = cvcuda.cvtcolor_into(
        src=input,
        dst=output,
        code=code,
        stream=stream,
    )
    assert tmp is output
    assert len(output) == len(input)
    assert output.capacity == input.capacity
    assert output.maxsize == input.maxsize



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
import xml.etree.ElementTree as ET
import os
import sys

outdir = Path(sys.argv[1])

if not os.path.exists(outdir):
    os.makedirs(outdir)

xmlRoot = sys.argv[2]

for i in os.listdir(xmlRoot):
    group_path = os.path.join(xmlRoot, i)
    if os.path.isfile(group_path) and "group__" in i:
        tree = ET.parse(group_path)
        root = tree.getroot()
        for compounddef in root.iter("compounddef"):
            group_name = compounddef.attrib["id"]
            group_label = compounddef.find("compoundname").text
            group_title = compounddef.find("title").text
            outfile = outdir / (group_name + ".rst")
            output = ":orphan:\n\n"
            output += group_title + "\n"
            output += "=" * len(group_title) + "\n\n"
            output += f".. doxygengroup:: {group_label}\n"
            output += "   :project: cvcuda\n"
            outfile.write_text(output)



import sys, os
import subprocess

def build(setuptools_kwargs):
    print(setuptools_kwargs)
    build_dir = "build-rel"#os.path.join(".", "build-rel")
    os.makedirs(build_dir)

    print("Initializing git submodules")
    subprocess.check_call("git submodule update --init".split(" "))

    print("Retrieving git LFS objects")
    subprocess.check_call("git lfs fetch".split(" "))
    subprocess.check_call("git lfs checkout".split(" "))

    print("Building to", build_dir)
    print("#"*10, "Please Be Patient", "#"*10)
    my_env = os.environ.copy()

    my_env['MAKE_OPTS'] = '-j4'
    subprocess.check_call(
        ["./ci/build.sh", "release", os.path.join("build-rel")],
        stdout=sys.stdout,
        stderr=subprocess.STDOUT,
        env=my_env,
        shell=True)
    #    subprocess.run("./ci/build.sh")
    #target_file = os.path.join(dst, "mypackage/myfile.txt")
    #os.makedirs(os.path.dirname(target_file), exist_ok=True)
    #os.makedirs(os.path.dirname(target_file), exist_ok=True)
    #download_file_to(dst)



#!/usr/bin/env python
#
# Copyright 2006, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Unit test for Google Test's --gtest_list_tests flag.

A user can ask Google Test to list all tests by specifying the
--gtest_list_tests flag. If output is requested, via --gtest_output=xml
or --gtest_output=json, the tests are listed, with extra information in the
output file.
This script tests such functionality by invoking gtest_list_output_unittest_
 (a program written with Google Test) the command line flags.
"""

import os
import re
from googletest.test import gtest_test_utils

GTEST_LIST_TESTS_FLAG = '--gtest_list_tests'
GTEST_OUTPUT_FLAG = '--gtest_output'

EXPECTED_XML = """<\?xml version="1.0" encoding="UTF-8"\?>
<testsuites tests="16" name="AllTests">
  <testsuite name="FooTest" tests="2">
    <testcase name="Test1" file=".*gtest_list_output_unittest_.cc" line="43" />
    <testcase name="Test2" file=".*gtest_list_output_unittest_.cc" line="45" />
  </testsuite>
  <testsuite name="FooTestFixture" tests="2">
    <testcase name="Test3" file=".*gtest_list_output_unittest_.cc" line="48" />
    <testcase name="Test4" file=".*gtest_list_output_unittest_.cc" line="49" />
  </testsuite>
  <testsuite name="TypedTest/0" tests="2">
    <testcase name="Test7" type_param="int" file=".*gtest_list_output_unittest_.cc" line="60" />
    <testcase name="Test8" type_param="int" file=".*gtest_list_output_unittest_.cc" line="61" />
  </testsuite>
  <testsuite name="TypedTest/1" tests="2">
    <testcase name="Test7" type_param="bool" file=".*gtest_list_output_unittest_.cc" line="60" />
    <testcase name="Test8" type_param="bool" file=".*gtest_list_output_unittest_.cc" line="61" />
  </testsuite>
  <testsuite name="Single/TypeParameterizedTestSuite/0" tests="2">
    <testcase name="Test9" type_param="int" file=".*gtest_list_output_unittest_.cc" line="66" />
    <testcase name="Test10" type_param="int" file=".*gtest_list_output_unittest_.cc" line="67" />
  </testsuite>
  <testsuite name="Single/TypeParameterizedTestSuite/1" tests="2">
    <testcase name="Test9" type_param="bool" file=".*gtest_list_output_unittest_.cc" line="66" />
    <testcase name="Test10" type_param="bool" file=".*gtest_list_output_unittest_.cc" line="67" />
  </testsuite>
  <testsuite name="ValueParam/ValueParamTest" tests="4">
    <testcase name="Test5/0" value_param="33" file=".*gtest_list_output_unittest_.cc" line="52" />
    <testcase name="Test5/1" value_param="42" file=".*gtest_list_output_unittest_.cc" line="52" />
    <testcase name="Test6/0" value_param="33" file=".*gtest_list_output_unittest_.cc" line="53" />
    <testcase name="Test6/1" value_param="42" file=".*gtest_list_output_unittest_.cc" line="53" />
  </testsuite>
</testsuites>
"""

EXPECTED_JSON = """{
  "tests": 16,
  "name": "AllTests",
  "testsuites": \[
    {
      "name": "FooTest",
      "tests": 2,
      "testsuite": \[
        {
          "name": "Test1",
          "file": ".*gtest_list_output_unittest_.cc",
          "line": 43
        },
        {
          "name": "Test2",
          "file": ".*gtest_list_output_unittest_.cc",
          "line": 45
        }
      \]
    },
    {
      "name": "FooTestFixture",
      "tests": 2,
      "testsuite": \[
        {
          "name": "Test3",
          "file": ".*gtest_list_output_unittest_.cc",
          "line": 48
        },
        {
          "name": "Test4",
          "file": ".*gtest_list_output_unittest_.cc",
          "line": 49
        }
      \]
    },
    {
      "name": "TypedTest\\\\/0",
      "tests": 2,
      "testsuite": \[
        {
          "name": "Test7",
          "type_param": "int",
          "file": ".*gtest_list_output_unittest_.cc",
          "line": 60
        },
        {
          "name": "Test8",
          "type_param": "int",
          "file": ".*gtest_list_output_unittest_.cc",
          "line": 61
        }
      \]
    },
    {
      "name": "TypedTest\\\\/1",
      "tests": 2,
      "testsuite": \[
        {
          "name": "Test7",
          "type_param": "bool",
          "file": ".*gtest_list_output_unittest_.cc",
          "line": 60
        },
        {
          "name": "Test8",
          "type_param": "bool",
          "file": ".*gtest_list_output_unittest_.cc",
          "line": 61
        }
      \]
    },
    {
      "name": "Single\\\\/TypeParameterizedTestSuite\\\\/0",
      "tests": 2,
      "testsuite": \[
        {
          "name": "Test9",
          "type_param": "int",
          "file": ".*gtest_list_output_unittest_.cc",
          "line": 66
        },
        {
          "name": "Test10",
          "type_param": "int",
          "file": ".*gtest_list_output_unittest_.cc",
          "line": 67
        }
      \]
    },
    {
      "name": "Single\\\\/TypeParameterizedTestSuite\\\\/1",
      "tests": 2,
      "testsuite": \[
        {
          "name": "Test9",
          "type_param": "bool",
          "file": ".*gtest_list_output_unittest_.cc",
          "line": 66
        },
        {
          "name": "Test10",
          "type_param": "bool",
          "file": ".*gtest_list_output_unittest_.cc",
          "line": 67
        }
      \]
    },
    {
      "name": "ValueParam\\\\/ValueParamTest",
      "tests": 4,
      "testsuite": \[
        {
          "name": "Test5\\\\/0",
          "value_param": "33",
          "file": ".*gtest_list_output_unittest_.cc",
          "line": 52
        },
        {
          "name": "Test5\\\\/1",
          "value_param": "42",
          "file": ".*gtest_list_output_unittest_.cc",
          "line": 52
        },
        {
          "name": "Test6\\\\/0",
          "value_param": "33",
          "file": ".*gtest_list_output_unittest_.cc",
          "line": 53
        },
        {
          "name": "Test6\\\\/1",
          "value_param": "42",
          "file": ".*gtest_list_output_unittest_.cc",
          "line": 53
        }
      \]
    }
  \]
}
"""


class GTestListTestsOutputUnitTest(gtest_test_utils.TestCase):
  """Unit test for Google Test's list tests with output to file functionality.
  """

  def testXml(self):
    """Verifies XML output for listing tests in a Google Test binary.

    Runs a test program that generates an empty XML output, and
    tests that the XML output is expected.
    """
    self._TestOutput('xml', EXPECTED_XML)

  def testJSON(self):
    """Verifies XML output for listing tests in a Google Test binary.

    Runs a test program that generates an empty XML output, and
    tests that the XML output is expected.
    """
    self._TestOutput('json', EXPECTED_JSON)

  def _GetOutput(self, out_format):
    file_path = os.path.join(gtest_test_utils.GetTempDir(),
                             'test_out.' + out_format)
    gtest_prog_path = gtest_test_utils.GetTestExecutablePath(
        'gtest_list_output_unittest_')

    command = ([
        gtest_prog_path,
        '%s=%s:%s' % (GTEST_OUTPUT_FLAG, out_format, file_path),
        '--gtest_list_tests'
    ])
    environ_copy = os.environ.copy()
    p = gtest_test_utils.Subprocess(
        command, env=environ_copy, working_dir=gtest_test_utils.GetTempDir())

    self.assertTrue(p.exited)
    self.assertEqual(0, p.exit_code)
    self.assertTrue(os.path.isfile(file_path))
    with open(file_path) as f:
      result = f.read()
    return result

  def _TestOutput(self, test_format, expected_output):
    actual = self._GetOutput(test_format)
    actual_lines = actual.splitlines()
    expected_lines = expected_output.splitlines()
    line_count = 0
    for actual_line in actual_lines:
      expected_line = expected_lines[line_count]
      expected_line_re = re.compile(expected_line.strip())
      self.assertTrue(
          expected_line_re.match(actual_line.strip()),
          ('actual output of "%s",\n'
           'which does not match expected regex of "%s"\n'
           'on line %d' % (actual, expected_output, line_count)))
      line_count = line_count + 1


if __name__ == '__main__':
  os.environ['GTEST_STACK_TRACE_DEPTH'] = '1'
  gtest_test_utils.Main()



import pytest

import env  # noqa: F401
from pybind11_tests import numpy_array as m

np = pytest.importorskip("numpy")


def test_dtypes():
    # See issue #1328.
    # - Platform-dependent sizes.
    for size_check in m.get_platform_dtype_size_checks():
        print(size_check)
        assert size_check.size_cpp == size_check.size_numpy, size_check
    # - Concrete sizes.
    for check in m.get_concrete_dtype_checks():
        print(check)
        assert check.numpy == check.pybind11, check
        if check.numpy.num != check.pybind11.num:
            print(
                f"NOTE: typenum mismatch for {check}: {check.numpy.num} != {check.pybind11.num}"
            )


@pytest.fixture(scope="function")
def arr():
    return np.array([[1, 2, 3], [4, 5, 6]], "=u2")


def test_array_attributes():
    a = np.array(0, "f8")
    assert m.ndim(a) == 0
    assert all(m.shape(a) == [])
    assert all(m.strides(a) == [])
    with pytest.raises(IndexError) as excinfo:
        m.shape(a, 0)
    assert str(excinfo.value) == "invalid axis: 0 (ndim = 0)"
    with pytest.raises(IndexError) as excinfo:
        m.strides(a, 0)
    assert str(excinfo.value) == "invalid axis: 0 (ndim = 0)"
    assert m.writeable(a)
    assert m.size(a) == 1
    assert m.itemsize(a) == 8
    assert m.nbytes(a) == 8
    assert m.owndata(a)

    a = np.array([[1, 2, 3], [4, 5, 6]], "u2").view()
    a.flags.writeable = False
    assert m.ndim(a) == 2
    assert all(m.shape(a) == [2, 3])
    assert m.shape(a, 0) == 2
    assert m.shape(a, 1) == 3
    assert all(m.strides(a) == [6, 2])
    assert m.strides(a, 0) == 6
    assert m.strides(a, 1) == 2
    with pytest.raises(IndexError) as excinfo:
        m.shape(a, 2)
    assert str(excinfo.value) == "invalid axis: 2 (ndim = 2)"
    with pytest.raises(IndexError) as excinfo:
        m.strides(a, 2)
    assert str(excinfo.value) == "invalid axis: 2 (ndim = 2)"
    assert not m.writeable(a)
    assert m.size(a) == 6
    assert m.itemsize(a) == 2
    assert m.nbytes(a) == 12
    assert not m.owndata(a)


@pytest.mark.parametrize(
    "args, ret", [([], 0), ([0], 0), ([1], 3), ([0, 1], 1), ([1, 2], 5)]
)
def test_index_offset(arr, args, ret):
    assert m.index_at(arr, *args) == ret
    assert m.index_at_t(arr, *args) == ret
    assert m.offset_at(arr, *args) == ret * arr.dtype.itemsize
    assert m.offset_at_t(arr, *args) == ret * arr.dtype.itemsize


def test_dim_check_fail(arr):
    for func in (
        m.index_at,
        m.index_at_t,
        m.offset_at,
        m.offset_at_t,
        m.data,
        m.data_t,
        m.mutate_data,
        m.mutate_data_t,
    ):
        with pytest.raises(IndexError) as excinfo:
            func(arr, 1, 2, 3)
        assert str(excinfo.value) == "too many indices for an array: 3 (ndim = 2)"


@pytest.mark.parametrize(
    "args, ret",
    [
        ([], [1, 2, 3, 4, 5, 6]),
        ([1], [4, 5, 6]),
        ([0, 1], [2, 3, 4, 5, 6]),
        ([1, 2], [6]),
    ],
)
def test_data(arr, args, ret):
    from sys import byteorder

    assert all(m.data_t(arr, *args) == ret)
    assert all(m.data(arr, *args)[(0 if byteorder == "little" else 1) :: 2] == ret)
    assert all(m.data(arr, *args)[(1 if byteorder == "little" else 0) :: 2] == 0)


@pytest.mark.parametrize("dim", [0, 1, 3])
def test_at_fail(arr, dim):
    for func in m.at_t, m.mutate_at_t:
        with pytest.raises(IndexError) as excinfo:
            func(arr, *([0] * dim))
        assert str(excinfo.value) == f"index dimension mismatch: {dim} (ndim = 2)"


def test_at(arr):
    assert m.at_t(arr, 0, 2) == 3
    assert m.at_t(arr, 1, 0) == 4

    assert all(m.mutate_at_t(arr, 0, 2).ravel() == [1, 2, 4, 4, 5, 6])
    assert all(m.mutate_at_t(arr, 1, 0).ravel() == [1, 2, 4, 5, 5, 6])


def test_mutate_readonly(arr):
    arr.flags.writeable = False
    for func, args in (
        (m.mutate_data, ()),
        (m.mutate_data_t, ()),
        (m.mutate_at_t, (0, 0)),
    ):
        with pytest.raises(ValueError) as excinfo:
            func(arr, *args)
        assert str(excinfo.value) == "array is not writeable"


def test_mutate_data(arr):
    assert all(m.mutate_data(arr).ravel() == [2, 4, 6, 8, 10, 12])
    assert all(m.mutate_data(arr).ravel() == [4, 8, 12, 16, 20, 24])
    assert all(m.mutate_data(arr, 1).ravel() == [4, 8, 12, 32, 40, 48])
    assert all(m.mutate_data(arr, 0, 1).ravel() == [4, 16, 24, 64, 80, 96])
    assert all(m.mutate_data(arr, 1, 2).ravel() == [4, 16, 24, 64, 80, 192])

    assert all(m.mutate_data_t(arr).ravel() == [5, 17, 25, 65, 81, 193])
    assert all(m.mutate_data_t(arr).ravel() == [6, 18, 26, 66, 82, 194])
    assert all(m.mutate_data_t(arr, 1).ravel() == [6, 18, 26, 67, 83, 195])
    assert all(m.mutate_data_t(arr, 0, 1).ravel() == [6, 19, 27, 68, 84, 196])
    assert all(m.mutate_data_t(arr, 1, 2).ravel() == [6, 19, 27, 68, 84, 197])


def test_bounds_check(arr):
    for func in (
        m.index_at,
        m.index_at_t,
        m.data,
        m.data_t,
        m.mutate_data,
        m.mutate_data_t,
        m.at_t,
        m.mutate_at_t,
    ):
        with pytest.raises(IndexError) as excinfo:
            func(arr, 2, 0)
        assert str(excinfo.value) == "index 2 is out of bounds for axis 0 with size 2"
        with pytest.raises(IndexError) as excinfo:
            func(arr, 0, 4)
        assert str(excinfo.value) == "index 4 is out of bounds for axis 1 with size 3"


def test_make_c_f_array():
    assert m.make_c_array().flags.c_contiguous
    assert not m.make_c_array().flags.f_contiguous
    assert m.make_f_array().flags.f_contiguous
    assert not m.make_f_array().flags.c_contiguous


def test_make_empty_shaped_array():
    m.make_empty_shaped_array()

    # empty shape means numpy scalar, PEP 3118
    assert m.scalar_int().ndim == 0
    assert m.scalar_int().shape == ()
    assert m.scalar_int() == 42


def test_wrap():
    def assert_references(a, b, base=None):
        if base is None:
            base = a
        assert a is not b
        assert a.__array_interface__["data"][0] == b.__array_interface__["data"][0]
        assert a.shape == b.shape
        assert a.strides == b.strides
        assert a.flags.c_contiguous == b.flags.c_contiguous
        assert a.flags.f_contiguous == b.flags.f_contiguous
        assert a.flags.writeable == b.flags.writeable
        assert a.flags.aligned == b.flags.aligned
        # 1.13 supported Python 3.6
        if tuple(int(x) for x in np.__version__.split(".")[:2]) >= (1, 14):
            assert a.flags.writebackifcopy == b.flags.writebackifcopy
        else:
            assert a.flags.updateifcopy == b.flags.updateifcopy
        assert np.all(a == b)
        assert not b.flags.owndata
        assert b.base is base
        if a.flags.writeable and a.ndim == 2:
            a[0, 0] = 1234
            assert b[0, 0] == 1234

    a1 = np.array([1, 2], dtype=np.int16)
    assert a1.flags.owndata and a1.base is None
    a2 = m.wrap(a1)
    assert_references(a1, a2)

    a1 = np.array([[1, 2], [3, 4]], dtype=np.float32, order="F")
    assert a1.flags.owndata and a1.base is None
    a2 = m.wrap(a1)
    assert_references(a1, a2)

    a1 = np.array([[1, 2], [3, 4]], dtype=np.float32, order="C")
    a1.flags.writeable = False
    a2 = m.wrap(a1)
    assert_references(a1, a2)

    a1 = np.random.random((4, 4, 4))
    a2 = m.wrap(a1)
    assert_references(a1, a2)

    a1t = a1.transpose()
    a2 = m.wrap(a1t)
    assert_references(a1t, a2, a1)

    a1d = a1.diagonal()
    a2 = m.wrap(a1d)
    assert_references(a1d, a2, a1)

    a1m = a1[::-1, ::-1, ::-1]
    a2 = m.wrap(a1m)
    assert_references(a1m, a2, a1)


def test_numpy_view(capture):
    with capture:
        ac = m.ArrayClass()
        ac_view_1 = ac.numpy_view()
        ac_view_2 = ac.numpy_view()
        assert np.all(ac_view_1 == np.array([1, 2], dtype=np.int32))
        del ac
        pytest.gc_collect()
    assert (
        capture
        == """
        ArrayClass()
        ArrayClass::numpy_view()
        ArrayClass::numpy_view()
    """
    )
    ac_view_1[0] = 4
    ac_view_1[1] = 3
    assert ac_view_2[0] == 4
    assert ac_view_2[1] == 3
    with capture:
        del ac_view_1
        del ac_view_2
        pytest.gc_collect()
        pytest.gc_collect()
    assert (
        capture
        == """
        ~ArrayClass()
    """
    )


def test_cast_numpy_int64_to_uint64():
    m.function_taking_uint64(123)
    m.function_taking_uint64(np.uint64(123))


def test_isinstance():
    assert m.isinstance_untyped(np.array([1, 2, 3]), "not an array")
    assert m.isinstance_typed(np.array([1.0, 2.0, 3.0]))


def test_constructors():
    defaults = m.default_constructors()
    for a in defaults.values():
        assert a.size == 0
    assert defaults["array"].dtype == np.array([]).dtype
    assert defaults["array_t<int32>"].dtype == np.int32
    assert defaults["array_t<double>"].dtype == np.float64

    results = m.converting_constructors([1, 2, 3])
    for a in results.values():
        np.testing.assert_array_equal(a, [1, 2, 3])
    assert results["array"].dtype == np.int_
    assert results["array_t<int32>"].dtype == np.int32
    assert results["array_t<double>"].dtype == np.float64


def test_overload_resolution(msg):
    # Exact overload matches:
    assert m.overloaded(np.array([1], dtype="float64")) == "double"
    assert m.overloaded(np.array([1], dtype="float32")) == "float"
    assert m.overloaded(np.array([1], dtype="ushort")) == "unsigned short"
    assert m.overloaded(np.array([1], dtype="intc")) == "int"
    assert m.overloaded(np.array([1], dtype="longlong")) == "long long"
    assert m.overloaded(np.array([1], dtype="complex")) == "double complex"
    assert m.overloaded(np.array([1], dtype="csingle")) == "float complex"

    # No exact match, should call first convertible version:
    assert m.overloaded(np.array([1], dtype="uint8")) == "double"

    with pytest.raises(TypeError) as excinfo:
        m.overloaded("not an array")
    assert (
        msg(excinfo.value)
        == """
        overloaded(): incompatible function arguments. The following argument types are supported:
            1. (arg0: numpy.ndarray[numpy.float64]) -> str
            2. (arg0: numpy.ndarray[numpy.float32]) -> str
            3. (arg0: numpy.ndarray[numpy.int32]) -> str
            4. (arg0: numpy.ndarray[numpy.uint16]) -> str
            5. (arg0: numpy.ndarray[numpy.int64]) -> str
            6. (arg0: numpy.ndarray[numpy.complex128]) -> str
            7. (arg0: numpy.ndarray[numpy.complex64]) -> str

        Invoked with: 'not an array'
    """
    )

    assert m.overloaded2(np.array([1], dtype="float64")) == "double"
    assert m.overloaded2(np.array([1], dtype="float32")) == "float"
    assert m.overloaded2(np.array([1], dtype="complex64")) == "float complex"
    assert m.overloaded2(np.array([1], dtype="complex128")) == "double complex"
    assert m.overloaded2(np.array([1], dtype="float32")) == "float"

    assert m.overloaded3(np.array([1], dtype="float64")) == "double"
    assert m.overloaded3(np.array([1], dtype="intc")) == "int"
    expected_exc = """
        overloaded3(): incompatible function arguments. The following argument types are supported:
            1. (arg0: numpy.ndarray[numpy.int32]) -> str
            2. (arg0: numpy.ndarray[numpy.float64]) -> str

        Invoked with: """

    with pytest.raises(TypeError) as excinfo:
        m.overloaded3(np.array([1], dtype="uintc"))
    assert msg(excinfo.value) == expected_exc + repr(np.array([1], dtype="uint32"))
    with pytest.raises(TypeError) as excinfo:
        m.overloaded3(np.array([1], dtype="float32"))
    assert msg(excinfo.value) == expected_exc + repr(np.array([1.0], dtype="float32"))
    with pytest.raises(TypeError) as excinfo:
        m.overloaded3(np.array([1], dtype="complex"))
    assert msg(excinfo.value) == expected_exc + repr(np.array([1.0 + 0.0j]))

    # Exact matches:
    assert m.overloaded4(np.array([1], dtype="double")) == "double"
    assert m.overloaded4(np.array([1], dtype="longlong")) == "long long"
    # Non-exact matches requiring conversion.  Since float to integer isn't a
    # save conversion, it should go to the double overload, but short can go to
    # either (and so should end up on the first-registered, the long long).
    assert m.overloaded4(np.array([1], dtype="float32")) == "double"
    assert m.overloaded4(np.array([1], dtype="short")) == "long long"

    assert m.overloaded5(np.array([1], dtype="double")) == "double"
    assert m.overloaded5(np.array([1], dtype="uintc")) == "unsigned int"
    assert m.overloaded5(np.array([1], dtype="float32")) == "unsigned int"


def test_greedy_string_overload():
    """Tests fix for #685 - ndarray shouldn't go to std::string overload"""

    assert m.issue685("abc") == "string"
    assert m.issue685(np.array([97, 98, 99], dtype="b")) == "array"
    assert m.issue685(123) == "other"


def test_array_unchecked_fixed_dims(msg):
    z1 = np.array([[1, 2], [3, 4]], dtype="float64")
    m.proxy_add2(z1, 10)
    assert np.all(z1 == [[11, 12], [13, 14]])

    with pytest.raises(ValueError) as excinfo:
        m.proxy_add2(np.array([1.0, 2, 3]), 5.0)
    assert (
        msg(excinfo.value) == "array has incorrect number of dimensions: 1; expected 2"
    )

    expect_c = np.ndarray(shape=(3, 3, 3), buffer=np.array(range(3, 30)), dtype="int")
    assert np.all(m.proxy_init3(3.0) == expect_c)
    expect_f = np.transpose(expect_c)
    assert np.all(m.proxy_init3F(3.0) == expect_f)

    assert m.proxy_squared_L2_norm(np.array(range(6))) == 55
    assert m.proxy_squared_L2_norm(np.array(range(6), dtype="float64")) == 55

    assert m.proxy_auxiliaries2(z1) == [11, 11, True, 2, 8, 2, 2, 4, 32]
    assert m.proxy_auxiliaries2(z1) == m.array_auxiliaries2(z1)

    assert m.proxy_auxiliaries1_const_ref(z1[0, :])
    assert m.proxy_auxiliaries2_const_ref(z1)


def test_array_unchecked_dyn_dims():
    z1 = np.array([[1, 2], [3, 4]], dtype="float64")
    m.proxy_add2_dyn(z1, 10)
    assert np.all(z1 == [[11, 12], [13, 14]])

    expect_c = np.ndarray(shape=(3, 3, 3), buffer=np.array(range(3, 30)), dtype="int")
    assert np.all(m.proxy_init3_dyn(3.0) == expect_c)

    assert m.proxy_auxiliaries2_dyn(z1) == [11, 11, True, 2, 8, 2, 2, 4, 32]
    assert m.proxy_auxiliaries2_dyn(z1) == m.array_auxiliaries2(z1)


def test_array_failure():
    with pytest.raises(ValueError) as excinfo:
        m.array_fail_test()
    assert str(excinfo.value) == "cannot create a pybind11::array from a nullptr"

    with pytest.raises(ValueError) as excinfo:
        m.array_t_fail_test()
    assert str(excinfo.value) == "cannot create a pybind11::array_t from a nullptr"

    with pytest.raises(ValueError) as excinfo:
        m.array_fail_test_negative_size()
    assert str(excinfo.value) == "negative dimensions are not allowed"


def test_initializer_list():
    assert m.array_initializer_list1().shape == (1,)
    assert m.array_initializer_list2().shape == (1, 2)
    assert m.array_initializer_list3().shape == (1, 2, 3)
    assert m.array_initializer_list4().shape == (1, 2, 3, 4)


def test_array_resize():
    a = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9], dtype="float64")
    m.array_reshape2(a)
    assert a.size == 9
    assert np.all(a == [[1, 2, 3], [4, 5, 6], [7, 8, 9]])

    # total size change should succced with refcheck off
    m.array_resize3(a, 4, False)
    assert a.size == 64
    # ... and fail with refcheck on
    try:
        m.array_resize3(a, 3, True)
    except ValueError as e:
        assert str(e).startswith("cannot resize an array")
    # transposed array doesn't own data
    b = a.transpose()
    try:
        m.array_resize3(b, 3, False)
    except ValueError as e:
        assert str(e).startswith("cannot resize this array: it does not own its data")
    # ... but reshape should be fine
    m.array_reshape2(b)
    assert b.shape == (8, 8)


@pytest.mark.xfail("env.PYPY")
def test_array_create_and_resize():
    a = m.create_and_resize(2)
    assert a.size == 4
    assert np.all(a == 42.0)


def test_array_view():
    a = np.ones(100 * 4).astype("uint8")
    a_float_view = m.array_view(a, "float32")
    assert a_float_view.shape == (100 * 1,)  # 1 / 4 bytes = 8 / 32

    a_int16_view = m.array_view(a, "int16")  # 1 / 2 bytes = 16 / 32
    assert a_int16_view.shape == (100 * 2,)


def test_array_view_invalid():
    a = np.ones(100 * 4).astype("uint8")
    with pytest.raises(TypeError):
        m.array_view(a, "deadly_dtype")


def test_reshape_initializer_list():
    a = np.arange(2 * 7 * 3) + 1
    x = m.reshape_initializer_list(a, 2, 7, 3)
    assert x.shape == (2, 7, 3)
    assert list(x[1][4]) == [34, 35, 36]
    with pytest.raises(ValueError) as excinfo:
        m.reshape_initializer_list(a, 1, 7, 3)
    assert str(excinfo.value) == "cannot reshape array of size 42 into shape (1,7,3)"


def test_reshape_tuple():
    a = np.arange(3 * 7 * 2) + 1
    x = m.reshape_tuple(a, (3, 7, 2))
    assert x.shape == (3, 7, 2)
    assert list(x[1][4]) == [23, 24]
    y = m.reshape_tuple(x, (x.size,))
    assert y.shape == (42,)
    with pytest.raises(ValueError) as excinfo:
        m.reshape_tuple(a, (3, 7, 1))
    assert str(excinfo.value) == "cannot reshape array of size 42 into shape (3,7,1)"
    with pytest.raises(ValueError) as excinfo:
        m.reshape_tuple(a, ())
    assert str(excinfo.value) == "cannot reshape array of size 42 into shape ()"


def test_index_using_ellipsis():
    a = m.index_using_ellipsis(np.zeros((5, 6, 7)))
    assert a.shape == (6,)


@pytest.mark.parametrize(
    "test_func",
    [
        m.test_fmt_desc_float,
        m.test_fmt_desc_double,
        m.test_fmt_desc_const_float,
        m.test_fmt_desc_const_double,
    ],
)
def test_format_descriptors_for_floating_point_types(test_func):
    assert "numpy.ndarray[numpy.float" in test_func.__doc__


@pytest.mark.parametrize("forcecast", [False, True])
@pytest.mark.parametrize("contiguity", [None, "C", "F"])
@pytest.mark.parametrize("noconvert", [False, True])
@pytest.mark.filterwarnings(
    "ignore:Casting complex values to real discards the imaginary part:numpy.ComplexWarning"
)
def test_argument_conversions(forcecast, contiguity, noconvert):
    function_name = "accept_double"
    if contiguity == "C":
        function_name += "_c_style"
    elif contiguity == "F":
        function_name += "_f_style"
    if forcecast:
        function_name += "_forcecast"
    if noconvert:
        function_name += "_noconvert"
    function = getattr(m, function_name)

    for dtype in [np.dtype("float32"), np.dtype("float64"), np.dtype("complex128")]:
        for order in ["C", "F"]:
            for shape in [(2, 2), (1, 3, 1, 1), (1, 1, 1), (0,)]:
                if not noconvert:
                    # If noconvert is not passed, only complex128 needs to be truncated and
                    # "cannot be safely obtained". So without `forcecast`, the argument shouldn't
                    # be accepted.
                    should_raise = dtype.name == "complex128" and not forcecast
                else:
                    # If noconvert is passed, only float64 and the matching order is accepted.
                    # If at most one dimension has a size greater than 1, the array is also
                    # trivially contiguous.
                    trivially_contiguous = sum(1 for d in shape if d > 1) <= 1
                    should_raise = dtype.name != "float64" or (
                        contiguity is not None
                        and contiguity != order
                        and not trivially_contiguous
                    )

                array = np.zeros(shape, dtype=dtype, order=order)
                if not should_raise:
                    function(array)
                else:
                    with pytest.raises(
                        TypeError, match="incompatible function arguments"
                    ):
                        function(array)


@pytest.mark.xfail("env.PYPY")
def test_dtype_refcount_leak():
    from sys import getrefcount

    dtype = np.dtype(np.float_)
    a = np.array([1], dtype=dtype)
    before = getrefcount(dtype)
    m.ndim(a)
    after = getrefcount(dtype)
    assert after == before


def test_round_trip_float():
    arr = np.zeros((), np.float64)
    arr[()] = 37.2
    assert m.round_trip_float(arr) == 37.2



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cvcuda
import pytest as t
import numpy as np
import util


RNG = np.random.default_rng(0)


@t.mark.parametrize(
    "input, kernel_size, kernel_anchor, border",
    [
        (
            cvcuda.Tensor([5, 16, 23, 4], np.uint8, "NHWC"),
            [3, 3],
            [1, 1],
            cvcuda.Border.CONSTANT,
        ),
        (
            cvcuda.Tensor([4, 4, 3], np.float32, "HWC"),
            [5, 5],
            [0, 0],
            cvcuda.Border.REPLICATE,
        ),
        (
            cvcuda.Tensor([3, 88, 13, 3], np.uint16, "NHWC"),
            [7, 7],
            [2, 2],
            cvcuda.Border.REFLECT,
        ),
        (
            cvcuda.Tensor([3, 4, 4], np.int32, "HWC"),
            [9, 9],
            [-1, -1],
            cvcuda.Border.WRAP,
        ),
        (
            cvcuda.Tensor([1, 2, 3, 4], np.int16, "NHWC"),
            [11, 11],
            [8, 8],
            cvcuda.Border.REFLECT101,
        ),
    ],
)
def test_op_averageblur(input, kernel_size, kernel_anchor, border):
    out = cvcuda.averageblur(input, kernel_size, kernel_anchor, border)
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == input.dtype

    stream = cvcuda.Stream()
    out = cvcuda.Tensor(input.shape, input.dtype, input.layout)
    tmp = cvcuda.averageblur_into(
        src=input,
        dst=out,
        kernel_size=kernel_size,
        kernel_anchor=kernel_anchor,
        border=border,
        stream=stream,
    )
    assert tmp is out
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == input.dtype


@t.mark.parametrize(
    "num_images, img_format, img_size, max_pixel, max_kernel_size, border",
    [
        (
            10,
            cvcuda.Format.RGB8,
            (123, 321),
            256,
            (3, 3),
            cvcuda.Border.CONSTANT,
        ),
        (
            7,
            cvcuda.Format.RGBf32,
            (62, 35),
            1.0,
            (5, 5),
            cvcuda.Border.REPLICATE,
        ),
        (
            1,
            cvcuda.Format.U8,
            (33, 48),
            123,
            (7, 7),
            cvcuda.Border.REFLECT,
        ),
        (
            13,
            cvcuda.Format.S16,
            (26, 52),
            1234,
            (9, 9),
            cvcuda.Border.WRAP,
        ),
        (
            6,
            cvcuda.Format.S32,
            (77, 42),
            123456,
            (11, 11),
            cvcuda.Border.REFLECT101,
        ),
    ],
)
def test_op_averageblurvarshape(
    num_images, img_format, img_size, max_pixel, max_kernel_size, border
):

    input = util.create_image_batch(
        num_images, img_format, size=img_size, max_random=max_pixel, rng=RNG
    )

    kernel_size = util.create_tensor(
        (num_images, 2),
        np.int32,
        "NC",
        max_random=max_kernel_size,
        rng=RNG,
        transform_dist=util.dist_odd,
    )

    kernel_anchor = util.create_tensor(
        (num_images, 2), np.int32, "NC", max_random=max_kernel_size, rng=RNG
    )

    out = cvcuda.averageblur(
        input,
        max_kernel_size,
        kernel_size,
        kernel_anchor,
        border,
    )
    assert len(out) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == input.uniqueformat
    assert out.maxsize == input.maxsize

    stream = cvcuda.Stream()
    out = util.clone_image_batch(input)
    tmp = cvcuda.averageblur_into(
        src=input,
        dst=out,
        max_kernel_size=max_kernel_size,
        kernel_size=kernel_size,
        kernel_anchor=kernel_anchor,
        border=border,
        stream=stream,
    )
    assert tmp is out
    assert len(out) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == input.uniqueformat
    assert out.maxsize == input.maxsize



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cvcuda
import pytest as t


@t.mark.parametrize(
    "input, erasing_area_num, random, seed",
    [
        (cvcuda.Tensor([1, 460, 640, 3], cvcuda.Type.U8, "NHWC"), 1, False, 0),
        (cvcuda.Tensor([5, 460, 640, 3], cvcuda.Type.U8, "NHWC"), 1, True, 1),
    ],
)
def test_op_erase(input, erasing_area_num, random, seed):

    parameter_shape = [erasing_area_num]
    anchor = cvcuda.Tensor(parameter_shape, cvcuda.Type._2S32, "N")
    erasing = cvcuda.Tensor(parameter_shape, cvcuda.Type._3S32, "N")
    imgIdx = cvcuda.Tensor(parameter_shape, cvcuda.Type.S32, "N")
    values = cvcuda.Tensor(parameter_shape, cvcuda.Type.F32, "N")

    out = cvcuda.erase(input, anchor, erasing, values, imgIdx)
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == input.dtype

    out = cvcuda.Tensor(input.shape, input.dtype, input.layout)
    tmp = cvcuda.erase_into(out, input, anchor, erasing, values, imgIdx)
    assert tmp is out

    stream = cvcuda.Stream()
    out = cvcuda.erase(
        src=input,
        anchor=anchor,
        erasing=erasing,
        values=values,
        imgIdx=imgIdx,
        random=random,
        seed=seed,
        stream=stream,
    )
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == input.dtype

    tmp = cvcuda.erase_into(
        src=input,
        dst=out,
        anchor=anchor,
        erasing=erasing,
        values=values,
        imgIdx=imgIdx,
        random=random,
        seed=seed,
        stream=stream,
    )
    assert tmp is out



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nvcv
import pytest as t


def test_recti_default():
    r = nvcv.RectI()
    assert r.x == 0
    assert r.y == 0
    assert r.width == 0
    assert r.height == 0


@t.mark.parametrize("x,y,w,h", [(0, 0, 0, 0), (10, -12, -45, 14)])
def test_recti_ctor(x, y, w, h):
    r = nvcv.RectI(x, y, w, h)
    assert r.x == x
    assert r.y == y
    assert r.width == w
    assert r.height == h

    r = nvcv.RectI(y=y, width=w, height=h, x=x)
    assert r.x == x
    assert r.y == y
    assert r.width == w
    assert r.height == h



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Import python modules

import os
import glob
import torch
from torchvision import models
import torchnvjpeg
import numpy as np
import argparse

# tag: Import CVCUDA module
import cvcuda

# tag: Classification Sample

"""
Image Classification python sample

The image classification sample uses Resnet50 based model trained on Imagenet
The sample app pipeline includes preprocessing, inference and post process stages
which takes as input a batch of images and returns the TopN classification results
of each image.

This sample gives an overview of the interoperability of pytorch with CVCUDA
tensors and operators
"""


class ClassificationSample:
    def __init__(
        self,
        input_path,
        labels_file,
        batch_size,
        target_img_height,
        target_img_width,
        device_id,
    ):
        self.input_path = input_path
        self.batch_size = batch_size
        self.target_img_height = target_img_height
        self.target_img_width = target_img_width
        self.device_id = device_id
        self.labels_file = labels_file

        # tag: Image Loading
        # Start by parsing the input_path expression first.
        if os.path.isfile(self.input_path):
            # Read the input image file.
            self.file_names = [self.input_path] * self.batch_size
            # Then create a dummy list with the data from the same file to simulate a
            # batch.
            self.data = [open(path, "rb").read() for path in self.file_names]

        elif os.path.isdir(self.input_path):
            # It is a directory. Grab all the images from it.
            self.file_names = glob.glob(os.path.join(self.input_path, "*.jpg"))
            self.data = [open(path, "rb").read() for path in self.file_names]
            print("Read a total of %d JPEG images." % len(self.data))

        else:
            print(
                "Input path not found. "
                "It is neither a valid JPEG file nor a directory: %s" % self.input_path
            )
            exit(1)

        # tag: Validate other inputs
        if not os.path.isfile(self.labels_file):
            print("Labels file not found: %s" % self.labels_file)
            exit(1)

        if self.batch_size <= 0:
            print("batch_size must be a value >=1.")
            exit(1)

        if self.target_img_height < 10:
            print("target_img_height must be a value >=10.")
            exit(1)

        if self.target_img_width < 10:
            print("target_img_width must be a value >=10.")
            exit(1)

    def run(self):
        # Runs the complete sample end-to-end.
        max_image_size = 1024 * 1024 * 3  # Maximum possible image size.

        # tag: NvJpeg Decoder
        # We will batchify the file_list and data_list based on the
        # batch size and start processing these batches one by one.
        file_name_batches = [
            self.file_names[i : i + self.batch_size]  # noqa: E203
            for i in range(0, len(self.file_names), self.batch_size)
        ]
        data_batches = [
            self.data[i : i + self.batch_size]  # noqa: E203
            for i in range(0, len(self.data), self.batch_size)
        ]
        batch_idx = 0

        for file_name_batch, data_batch in zip(file_name_batches, data_batches):
            effective_batch_size = len(file_name_batch)
            print("Processing batch %d of %d" % (batch_idx + 1, len(file_name_batches)))

            # Decode in batch using torchnvjpeg decoder on the GPU.
            decoder = torchnvjpeg.Decoder(
                0,
                0,
                True,
                self.device_id,
                effective_batch_size,
                8,  # this is max_cpu_threads parameter. Not used internally.
                max_image_size,
                torch.cuda.current_stream(self.device_id),
            )
            image_tensor_list = decoder.batch_decode(data_batch)

            # Convert the list of tensors to a tensor itself.
            image_tensors = torch.stack(image_tensor_list)

            # tag: Wrapping into Tensor
            # A torch tensor can be wrapped into a CVCUDA Object using the "as_tensor"
            # function in the specified layout. The datatype and dimensions are derived
            # directly from the torch tensor.
            cvcuda_input_tensor = cvcuda.as_tensor(image_tensors, "NHWC")

            # tag: Preprocess
            """
            Preprocessing includes the following sequence of operations.
            Resize -> DataType Convert(U8->F32) -> Normalize
            (Apply mean and std deviation) -> Interleaved to Planar
            """

            # Resize
            # Resize to the input network dimensions
            cvcuda_resize_tensor = cvcuda.resize(
                cvcuda_input_tensor,
                (
                    effective_batch_size,
                    self.target_img_height,
                    self.target_img_width,
                    3,
                ),
                cvcuda.Interp.LINEAR,
            )

            # Convert to the data type and range of values needed by the input layer
            # i.e uint8->float. A Scale is applied to normalize the values in the
            # range 0-1
            cvcuda_convert_tensor = cvcuda.convertto(
                cvcuda_resize_tensor, np.float32, scale=1 / 255
            )

            """
            The input to the network needs to be normalized based on the mean and
            std deviation value to standardize the input data.
            """

            # Create a torch tensor to store the mean and standard deviation
            # values for R,G,B
            mean = [0.485, 0.456, 0.406]
            std = [0.229, 0.224, 0.225]
            base_tensor = torch.Tensor(mean)
            stddev_tensor = torch.Tensor(std)

            # Reshape the the number of channels. The R,G,B values scale and offset
            # will be applied to every color plane respectively across the batch
            base_tensor = torch.reshape(base_tensor, (1, 1, 1, 3)).cuda()
            stddev_tensor = torch.reshape(stddev_tensor, (1, 1, 1, 3)).cuda()

            # Wrap the torch tensor in a CVCUDA Tensor
            cvcuda_base_tensor = cvcuda.as_tensor(base_tensor, "NHWC")
            cvcuda_scale_tensor = cvcuda.as_tensor(stddev_tensor, "NHWC")

            # Apply the normalize operator and indicate the scale values are
            # std deviation i.e scale = 1/stddev
            cvcuda_norm_tensor = cvcuda.normalize(
                cvcuda_convert_tensor,
                base=cvcuda_base_tensor,
                scale=cvcuda_scale_tensor,
                flags=cvcuda.NormalizeFlags.SCALE_IS_STDDEV,
            )

            # The final stage in the preprocess pipeline includes converting the RGB
            # buffer into a planar buffer
            cvcuda_preprocessed_tensor = cvcuda.reformat(cvcuda_norm_tensor, "NCHW")

            # tag: Inference
            # Inference uses pytorch to run a resnet50 model on the preprocessed
            # input and outputs the classification scores for 1000 classes
            # Load Resnet model pretrained on Imagenet
            resnet50 = models.resnet50(pretrained=True)
            resnet50.to("cuda")
            resnet50.eval()

            # Run inference on the preprocessed input
            torch_preprocessed_tensor = torch.as_tensor(
                cvcuda_preprocessed_tensor.cuda(), device="cuda"
            )

            with torch.no_grad():
                infer_output = resnet50(torch_preprocessed_tensor)

            # tag: Postprocess
            """
            Postprocessing function normalizes the classification score from the network
            and sorts the scores to get the TopN classification scores.
            """
            # Apply softmax to Normalize scores between 0-1
            scores = torch.nn.functional.softmax(infer_output, dim=1)

            # Sort output scores in descending order
            _, indices = torch.sort(infer_output, descending=True)

            # tag: Display Top N Results
            # Read and parse the classes
            with open(self.labels_file, "r") as f:
                classes = [line.strip() for line in f.readlines()]

            # top results to print out
            topN = 5
            for img_idx in range(effective_batch_size):
                print(
                    "Result for the image: %d of %d"
                    % (img_idx + 1, effective_batch_size)
                )

                # Display Top N Results
                for idx in indices[img_idx][:topN]:
                    idx = idx.item()
                    print(
                        "\tClass : ",
                        classes[idx],
                        " Score : ",
                        scores[img_idx][idx].item(),
                    )


def main():
    parser = argparse.ArgumentParser(
        "Classification sample using CV-CUDA.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "-i",
        "--input_path",
        default="tabby_tiger_cat.jpg",
        type=str,
        help="Either a path to a JPEG image or a directory containing JPEG "
        "images to use as input.",
    )

    parser.add_argument(
        "-l",
        "--labels_file",
        default="imagenet-classes.txt",
        type=str,
        help="The labels file to read and parse.",
    )

    parser.add_argument(
        "-th",
        "--target_img_height",
        default=224,
        type=int,
        help="The height to which you want to resize the input_image before "
        "running inference.",
    )

    parser.add_argument(
        "-tw",
        "--target_img_width",
        default=224,
        type=int,
        help="The width to which you want to resize the input_image before "
        "running inference.",
    )

    parser.add_argument(
        "-b", "--batch_size", default=1, type=int, help="Input Batch size"
    )

    parser.add_argument(
        "-d",
        "--device_id",
        default=0,
        type=int,
        help="The GPU device to use for this sample.",
    )

    # Parse the command line arguments.
    args = parser.parse_args()

    # Run the sample.
    sample = ClassificationSample(
        args.input_path,
        args.labels_file,
        args.batch_size,
        args.target_img_height,
        args.target_img_width,
        args.device_id,
    )

    sample.run()


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nvcv


def test_imgbatchvarshape_creation_works():
    batch = nvcv.ImageBatchVarShape(15)
    assert batch.capacity == 15
    assert len(batch) == 0
    assert batch.uniqueformat is None
    assert batch.maxsize == (0, 0)

    # range must be empty
    cnt = 0
    for i in batch:
        cnt += 1
    assert cnt == 0


def test_imgbatchvarshape_one_image():
    batch = nvcv.ImageBatchVarShape(15)

    img = nvcv.Image((64, 32), nvcv.Format.RGBA8)
    batch.pushback(img)
    assert len(batch) == 1
    assert batch.uniqueformat == nvcv.Format.RGBA8
    assert batch.maxsize == (64, 32)

    # range must contain one
    cnt = 0
    for bimg in batch:
        assert bimg is img
        cnt += 1
    assert cnt == 1

    # remove added image
    batch.popback()

    # check if its indeed removed
    assert len(batch) == 0
    cnt = 0
    for bimg in batch:
        cnt += 1
    assert cnt == 0


def test_imgbatchvarshape_several_images():
    batch = nvcv.ImageBatchVarShape(15)

    # add 4 images with different dimensions
    imgs = [nvcv.Image((m * 2, m), nvcv.Format.RGBA8) for m in range(2, 10, 2)]
    batch.pushback(imgs)
    assert len(batch) == 4
    assert batch.maxsize == (16, 8)
    assert batch.uniqueformat == nvcv.Format.RGBA8

    # check if they were really added
    cnt = 0
    for bimg in batch:
        assert bimg is imgs[cnt]
        cnt += 1
    assert cnt == 4

    # now remove the last 2
    batch.popback(2)
    assert len(batch) == 2
    cnt = 0
    for bimg in batch:
        cnt += 1
    assert cnt == 2

    assert batch.maxsize == (8, 4)

    # add one with a different format
    batch.pushback(nvcv.Image((58, 26), nvcv.Format.NV12))
    assert batch.maxsize == (58, 26)
    assert batch.uniqueformat is None

    # clear everything
    batch.clear()
    assert len(batch) == 0
    cnt = 0
    for bimg in batch:
        cnt += 1
    assert cnt == 0

    assert batch.maxsize == (0, 0)



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nvcv
import pytest as t
import numba
import numpy as np
from numba import cuda
import torch

assert numba.cuda.is_available()


@t.mark.parametrize(
    "n,size,fmt,gold_layout,gold_shape,gold_dtype",
    [
        (
            5,
            (32, 16),
            nvcv.Format.RGBA8,
            nvcv.TensorLayout.NHWC,
            [5, 16, 32, 4],
            np.uint8,
        ),
        (
            2,
            (38, 7),
            nvcv.Format.RGB8p,
            nvcv.TensorLayout.NCHW,
            [2, 3, 7, 38],
            np.uint8,
        ),
    ],
)
def test_tensor_creation_imagebatch_works(
    n, size, fmt, gold_layout, gold_shape, gold_dtype
):
    tensor = nvcv.Tensor(n, size, fmt)
    assert tensor.shape == gold_shape
    assert tensor.layout == gold_layout
    assert tensor.dtype == gold_dtype
    assert tensor.ndim == len(gold_shape)

    tensor = nvcv.Tensor(nimages=n, imgsize=size, format=fmt)
    assert tensor.shape == gold_shape
    assert tensor.layout == gold_layout
    assert tensor.dtype == gold_dtype
    assert tensor.ndim == len(gold_shape)


@t.mark.parametrize(
    "shape, dtype,layout",
    [
        ([5, 16, 32, 4], np.float32, nvcv.TensorLayout.NHWC),
        ([7, 3, 33, 11], np.complex64, nvcv.TensorLayout.NCHW),
        ([3, 11], np.int16, None),
        ([16, 32, 4], np.float32, nvcv.TensorLayout.HWC),
        ([32, 4], np.float32, nvcv.TensorLayout.WC),
        ([4, 32], np.float32, nvcv.TensorLayout.CW),
        ([32], np.float32, nvcv.TensorLayout.W),
    ],
)
def test_tensor_creation_shape_works(shape, dtype, layout):
    tensor = nvcv.Tensor(shape, dtype, layout)
    assert tensor.shape == shape
    assert tensor.dtype == dtype
    assert tensor.layout == layout
    assert tensor.ndim == len(shape)

    tensor = nvcv.Tensor(layout=layout, shape=shape, dtype=dtype)
    assert tensor.layout == layout
    assert tensor.dtype == dtype
    assert tensor.shape == shape
    assert tensor.ndim == len(shape)


@t.mark.parametrize(
    "shape,dtype",
    [
        ([3, 5, 7, 1], np.uint8),
        ([3, 5, 7, 1], np.int8),
        ([3, 5, 7, 1], np.uint16),
        ([3, 5, 7, 1], np.int16),
        ([3, 5, 7, 1], np.float32),
        ([3, 5, 7, 1], np.float64),
        ([3, 5, 7, 2], np.float32),
        ([3, 5, 7, 3], np.uint8),
        ([3, 5, 7, 4], np.uint8),
        ([3, 5, 7], np.csingle),
        ([3], np.int8),
    ],
)
def test_wrap_numba_buffer(shape, dtype):
    tensor = nvcv.as_tensor(cuda.device_array(shape, dtype))
    assert tensor.shape == shape
    assert tensor.dtype == dtype
    assert tensor.layout is None
    assert tensor.ndim == len(shape)


@t.mark.parametrize(
    "shape,dtype,layout",
    [
        ([3, 5, 7, 1], np.uint8, "NHWC"),
        ([3, 5, 7], np.uint8, "HWC"),
        ([3, 5, 7, 2], np.int16, "NHWC"),
        ([3, 5, 7, 2, 4, 2, 5], np.int16, "abcdefg"),
        ([3, 5], np.uint8, "HW"),
        ([5], np.uint8, "W"),
    ],
)
def test_wrap_numba_buffer_with_layout(shape, dtype, layout):
    tensor = nvcv.as_tensor(cuda.device_array(shape, dtype), layout)
    assert tensor.shape == shape
    assert tensor.dtype == dtype
    assert tensor.layout == layout
    assert tensor.ndim == len(shape)


@t.mark.parametrize(
    "size, fmt, gold_layout,gold_shape,gold_dtype",
    [
        (
            (32, 16),
            nvcv.Format.RGBA8,
            nvcv.TensorLayout.NHWC,
            [1, 16, 32, 4],
            np.uint8,
        ),
        (
            (38, 7),
            nvcv.Format.RGB8p,
            nvcv.TensorLayout.NCHW,
            [1, 3, 7, 38],
            np.uint8,
        ),
    ],
)
def test_tensor_wrap_image_works(size, fmt, gold_layout, gold_shape, gold_dtype):
    img = nvcv.Image(size, fmt)

    tensor = nvcv.as_tensor(img)
    assert tensor.shape == gold_shape
    assert tensor.layout == gold_layout
    assert tensor.dtype == gold_dtype


@t.mark.parametrize(
    "shape,dtype",
    [
        ([1, 23, 65, 3], np.uint8),
        ([5, 23, 65, 3], np.int8),
        ([65, 3], np.int16),
        ([243, 65, 3], np.uint16),
        ([1, 1], np.uint16),
        ([10], np.uint8),
    ],
)
def test_tensor_export_cuda_buffer(shape, dtype):

    hostGold = np.random.randint(0, 127, shape, dtype)

    devGold = cuda.to_device(hostGold)

    tensor = nvcv.as_tensor(devGold)

    devMem = tensor.cuda()
    assert devMem.dtype == dtype
    assert devMem.shape == shape

    assert (hostGold == cuda.as_cuda_array(devMem).copy_to_host()).all()


def test_tensor_hold_reference_of_wrapped_buffer():
    ttensor = torch.as_tensor(np.ndarray([10], np.int8), device="cuda")
    ptr0 = ttensor.data_ptr()

    cvtensor = nvcv.as_tensor(ttensor)  # noqa: F841 assigned but never used

    del ttensor  # cvtensor must have held ttensor object

    ttensor = torch.as_tensor(np.ndarray([10], np.int8), device="cuda")

    # since "cvtensor" must have held the reference to the first "ttensor",
    # the second "ttensor" must be a different buffer
    assert ptr0 != ttensor.data_ptr()



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nvcv
import torch
import numba
from numba import cuda
import pytest as t


assert cuda.is_available()


def test_current_stream():
    assert nvcv.cuda.Stream.current is nvcv.cuda.Stream.default
    assert type(nvcv.cuda.Stream.current) == nvcv.cuda.Stream


def test_user_stream():
    with nvcv.cuda.Stream():
        assert nvcv.cuda.Stream.current is not nvcv.cuda.Stream.default
    stream = nvcv.cuda.Stream()
    with stream:
        assert stream is nvcv.cuda.Stream.current
        assert stream is not nvcv.cuda.Stream.default
    assert stream is not nvcv.cuda.Stream.default
    assert stream is not nvcv.cuda.Stream.current


def test_nested_streams():
    stream1 = nvcv.cuda.Stream()
    stream2 = nvcv.cuda.Stream()
    assert stream1 is not stream2
    with stream1:
        with stream2:
            assert stream2 is nvcv.cuda.Stream.current
            assert stream1 is not nvcv.cuda.Stream.current
        assert stream2 is not nvcv.cuda.Stream.current
        assert stream1 is nvcv.cuda.Stream.current


def test_wrap_stream_voidp():
    numbaStream = numba.cuda.stream()
    nvcvStream = nvcv.cuda.as_stream(numbaStream.handle)

    assert numbaStream.handle.value == nvcvStream.handle


def test_wrap_stream_int():
    numbaStream = numba.cuda.stream()
    nvcvStream = nvcv.cuda.as_stream(numbaStream.handle.value)

    assert numbaStream.handle.value == nvcvStream.handle


def test_stream_conv_to_int():
    stream = nvcv.cuda.Stream()

    assert stream.handle == int(stream)


class NumbaStream:
    def __init__(self, cuda_stream=None):
        if cuda_stream:
            self.m_stream = numba.cuda.external_stream(cuda_stream)
        else:
            self.m_stream = numba.cuda.stream()

    def cuda_stream(self):
        return self.m_stream.handle.value

    def stream(self):
        return self.m_stream


class TorchStream:
    def __init__(self, cuda_stream=None):
        if cuda_stream:
            self.m_stream = torch.cuda.ExternalStream(cuda_stream)
        else:
            self.m_stream = torch.cuda.Stream()

    def cuda_stream(self):
        return self.m_stream.cuda_stream

    def stream(self):
        return self.m_stream


@t.mark.parametrize(
    "stream_type",
    [
        NumbaStream,
        TorchStream,
    ],
)
def test_wrap_stream_external(stream_type):
    extstream = stream_type()

    stream = nvcv.cuda.as_stream(extstream.stream())

    assert extstream.cuda_stream() == stream.handle

    # stream must hold a ref to the external stream, the wrapped cudaStream
    # must not have been deleted
    del extstream

    extstream = stream_type(stream.handle)
    stream = nvcv.cuda.as_stream(extstream.stream())

    assert extstream.cuda_stream() == stream.handle



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cvcuda
import pytest as t
import numpy as np
import util


@t.mark.parametrize(
    "input,out_shape,interp",
    [
        (
            cvcuda.Tensor([5, 16, 23, 4], np.uint8, "NHWC"),
            [5, 132, 15, 4],
            cvcuda.Interp.LINEAR,
        ),
        (
            cvcuda.Tensor([16, 23, 4], np.uint8, "HWC"),
            [132, 15, 4],
            cvcuda.Interp.CUBIC,
        ),
        (cvcuda.Tensor([16, 23, 1], np.uint8, "HWC"), [132, 15, 1], None),
    ],
)
def test_op_resize(input, out_shape, interp):
    if interp is None:
        out = cvcuda.resize(input, out_shape)
    else:
        out = cvcuda.resize(input, out_shape, interp)
    assert out.layout == input.layout
    assert out.shape == out_shape
    assert out.dtype == input.dtype

    out = cvcuda.Tensor(out_shape, input.dtype, input.layout)
    if interp is None:
        tmp = cvcuda.resize_into(out, input)
    else:
        tmp = cvcuda.resize_into(out, input, interp)
    assert tmp is out
    assert out.layout == input.layout
    assert out.shape == out_shape
    assert out.dtype == input.dtype

    stream = cvcuda.Stream()
    if interp is None:
        out = cvcuda.resize(src=input, shape=out_shape, stream=stream)
    else:
        out = cvcuda.resize(src=input, shape=out_shape, interp=interp, stream=stream)
    assert out.layout == input.layout
    assert out.shape == out_shape
    assert out.dtype == input.dtype

    if interp is None:
        tmp = cvcuda.resize_into(src=input, dst=out, stream=stream)
    else:
        tmp = cvcuda.resize_into(src=input, dst=out, interp=interp, stream=stream)
    assert tmp is out
    assert out.layout == input.layout
    assert out.shape == out_shape
    assert out.dtype == input.dtype


@t.mark.parametrize(
    "inSize, outSize, interp",
    [((123, 321), (321, 123), cvcuda.Interp.LINEAR), ((123, 321), (321, 123), None)],
)
def test_op_resizevarshape(inSize, outSize, interp):

    RNG = np.random.default_rng(0)

    input = util.create_image_batch(
        10, cvcuda.Format.RGBA8, size=inSize, max_random=256, rng=RNG
    )

    base_output = util.create_image_batch(
        10, cvcuda.Format.RGBA8, size=outSize, max_random=256, rng=RNG
    )

    sizes = []
    for image in base_output:
        sizes.append([image.width, image.height])

    if interp is None:
        out = cvcuda.resize(input, sizes)
    else:
        out = cvcuda.resize(src=input, sizes=sizes, interp=interp)

    assert len(out) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == input.uniqueformat
    assert out.maxsize == outSize

    stream = cvcuda.cuda.Stream()
    if interp is None:
        tmp = cvcuda.resize_into(src=input, dst=base_output, stream=stream)
    else:
        tmp = cvcuda.resize_into(
            src=input, dst=base_output, interp=interp, stream=stream
        )
    assert tmp is base_output
    assert len(base_output) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == input.uniqueformat
    assert out.maxsize == outSize



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nvcv
import pytest as t


@t.mark.parametrize(
    "format,gold_channels",
    [
        (nvcv.Format.RGBA8, 4),
        (nvcv.Format.RGB8, 3),
        (nvcv.Format._2S16, 2),
        (nvcv.Format.S8, 1),
        (nvcv.Format.NV12, 3),
    ],
)
def test_imgformat_numchannels(format, gold_channels):
    assert format.channels == gold_channels


@t.mark.parametrize(
    "format,gold_planes",
    [
        (nvcv.Format.RGBA8, 1),
        (nvcv.Format.RGB8, 1),
        (nvcv.Format._2S16, 1),
        (nvcv.Format.S8, 1),
        (nvcv.Format.NV12, 2),
    ],
)
def test_imgformat_planes(format, gold_planes):
    assert format.planes == gold_planes



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cvcuda
import pytest as t
import numpy as np
import util


RNG = np.random.default_rng(0)


@t.mark.parametrize(
    "input, xform, flags, border_mode, border_value",
    [
        (
            cvcuda.Tensor([5, 16, 23, 4], np.uint8, "NHWC"),
            np.array([[1, 0, 0], [0, 1, 0]]),
            cvcuda.Interp.NEAREST,
            cvcuda.Border.CONSTANT,
            [],
        ),
        (
            cvcuda.Tensor([5, 16, 23, 4], np.uint8, "NHWC"),
            np.array([[1, 0, 0], [0, 1, 0]]),
            cvcuda.Interp.NEAREST,
            cvcuda.Border.CONSTANT,
            [0],
        ),
        (
            cvcuda.Tensor([5, 16, 23, 4], np.uint8, "NHWC"),
            np.array([[1, 2, 0], [2, 1, 1]]),
            cvcuda.Interp.LINEAR,
            cvcuda.Border.WRAP,
            [1, 2, 3, 4],
        ),
        (
            cvcuda.Tensor([5, 16, 23, 4], np.uint8, "NHWC"),
            np.array([[1, 2, 0], [2, 1, 1]]),
            cvcuda.Interp.LINEAR,
            cvcuda.Border.REPLICATE,
            [1, 2, 3, 4],
        ),
        (
            cvcuda.Tensor([11, 21, 4], np.uint8, "HWC"),
            np.array([[2, 2, 0], [3, 1, 0]]),
            cvcuda.Interp.NEAREST,
            cvcuda.Border.CONSTANT,
            [0],
        ),
        (
            cvcuda.Tensor([11, 21, 4], np.uint8, "HWC"),
            np.array([[2, 2, 1], [3, 1, 2]]),
            cvcuda.Interp.LINEAR,
            cvcuda.Border.WRAP,
            [1, 2, 3, 4],
        ),
        (
            cvcuda.Tensor([11, 21, 4], np.uint8, "HWC"),
            np.array([[1, 2, 0], [2, 1, 1]]),
            cvcuda.Interp.LINEAR,
            cvcuda.Border.REPLICATE,
            [1, 2, 3, 4],
        ),
    ],
)
def test_op_warp_affine(input, xform, flags, border_mode, border_value):
    out = cvcuda.warp_affine(
        input, xform, flags, border_mode=border_mode, border_value=border_value
    )
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == input.dtype

    stream = cvcuda.Stream()
    out = cvcuda.Tensor(input.shape, input.dtype, input.layout)
    tmp = cvcuda.warp_affine_into(
        src=input,
        dst=out,
        xform=xform,
        flags=flags,
        border_mode=border_mode,
        border_value=border_value,
        stream=stream,
    )
    assert tmp is out
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == input.dtype


@t.mark.parametrize(
    "nimages, format, max_size, max_pixel, max_xval, flags, bmode, border_value",
    [
        (
            5,
            cvcuda.Format.RGB8,
            (16, 23),
            128.0,
            7,
            cvcuda.Interp.NEAREST,
            cvcuda.Border.CONSTANT,
            [],
        ),
        (
            5,
            cvcuda.Format.RGB8,
            (16, 23),
            128.0,
            7,
            cvcuda.Interp.NEAREST,
            cvcuda.Border.CONSTANT,
            [1, 2, 3, 4],
        ),
        (
            4,
            cvcuda.Format.RGB8,
            (16, 23),
            128.0,
            5,
            cvcuda.Interp.LINEAR,
            cvcuda.Border.WRAP,
            [0],
        ),
        (
            3,
            cvcuda.Format.RGB8,
            (16, 23),
            128.0,
            4,
            cvcuda.Interp.CUBIC,
            cvcuda.Border.REPLICATE,
            [2, 1, 0],
        ),
    ],
)
def test_op_warp_affinevarshape(
    nimages, format, max_size, max_pixel, max_xval, flags, bmode, border_value
):

    input = util.create_image_batch(
        nimages, format, max_size=max_size, max_random=max_pixel, rng=RNG
    )

    xform = util.create_tensor(
        (nimages, 6), np.float32, "NC", max_random=max_xval, rng=RNG
    )

    out = cvcuda.warp_affine(
        input, xform, flags, border_mode=bmode, border_value=border_value
    )
    assert len(out) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == input.uniqueformat
    assert out.maxsize == input.maxsize

    stream = cvcuda.Stream()

    out = util.clone_image_batch(input)
    tmp = cvcuda.warp_affine_into(
        src=input,
        dst=out,
        xform=xform,
        flags=flags,
        border_mode=bmode,
        border_value=border_value,
        stream=stream,
    )
    assert tmp is out
    assert len(out) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == input.uniqueformat
    assert out.maxsize == input.maxsize



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cvcuda
import pytest as t
import numpy as np


@t.mark.parametrize(
    "input,rc,out_shape",
    [
        (
            cvcuda.Tensor([5, 16, 23, 4], np.uint8, "NHWC"),
            cvcuda.RectI(x=0, y=0, width=20, height=2),
            [5, 2, 20, 4],
        ),
        (
            cvcuda.Tensor([5, 16, 23, 4], np.uint8, "NHWC"),
            cvcuda.RectI(2, 13, 16, 2),
            [5, 2, 16, 4],
        ),
        (
            cvcuda.Tensor([16, 23, 2], np.uint8, "HWC"),
            cvcuda.RectI(2, 13, 16, 2),
            [2, 16, 2],
        ),
    ],
)
def test_op_customcrop(input, rc, out_shape):
    out = cvcuda.customcrop(input, rc)
    assert out.layout == input.layout
    assert out.shape == out_shape
    assert out.dtype == input.dtype

    out = cvcuda.Tensor(input.shape, input.dtype, input.layout)
    tmp = cvcuda.customcrop_into(out, input, rc)
    assert tmp is out
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == input.dtype

    stream = cvcuda.Stream()
    out = cvcuda.customcrop(input, rect=rc, stream=stream)
    assert out.layout == input.layout
    assert out.shape == out_shape
    assert out.dtype == input.dtype

    out = cvcuda.Tensor(input.shape, input.dtype, input.layout)
    tmp = cvcuda.customcrop_into(dst=out, src=input, rect=rc, stream=stream)
    assert tmp is out
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == input.dtype



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cvcuda
import pytest as t
import numpy as np


@t.mark.parametrize(
    "input, crop_size, gold_shape",
    [
        (cvcuda.Tensor([5, 9, 9, 4], np.uint8, "NHWC"), [5, 5], [5, 5, 5, 4]),
        (cvcuda.Tensor([9, 9, 3], np.uint8, "HWC"), [5, 5], [5, 5, 3]),
        (cvcuda.Tensor([5, 21, 21, 4], np.uint8, "NHWC"), [15, 15], [5, 15, 15, 4]),
        (cvcuda.Tensor([21, 21, 3], np.uint8, "HWC"), [15, 15], [15, 15, 3]),
    ],
)
def test_op_center_crop(input, crop_size, gold_shape):
    out = cvcuda.center_crop(input, crop_size)
    assert out.layout == input.layout
    assert out.shape == gold_shape
    assert out.dtype == input.dtype

    stream = cvcuda.Stream()
    out = cvcuda.Tensor(input.shape, input.dtype, input.layout)
    tmp = cvcuda.center_crop_into(
        src=input,
        dst=out,
        crop_size=crop_size,
        stream=stream,
    )
    assert tmp is out
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == input.dtype



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#


# -- Project information -----------------------------------------------------
import os
import sphinx_rtd_theme

project = "CV-CUDA"
copyright = "2022, NVIDIA."
author = "NVIDIA"
version = "Alpha"
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = []

# Add any paths that contain templates here, relative to this directory.
templates_path = ["templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "docs/manuals/py/**"]

# source_parsers = { '.md': 'recommonmark.parser.CommonMarkParser',}

extensions = ["recommonmark"]


source_suffix = {".rst": "restructuredtext", ".md": "markdown"}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

html_theme = "sphinx_rtd_theme"
html_logo = os.path.join("content", "nv_logo.png")

html_theme_options = {
    "logo_only": True,
    "display_version": True,
    "prev_next_buttons_location": "bottom",
    "style_external_links": False,
    "style_nav_header_background": "#000000",
    # Toc options
    "collapse_navigation": False,
    "sticky_navigation": False,
    # 'navigation_depth': 10,
    "sidebarwidth": 12,
    "includehidden": True,
    "titles_only": False,
}


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".

html_favicon = os.path.join("content", "nv_icon.png")

html_static_path = ["templates"]

html_last_updated_fmt = ""

html_js_files = [
    "pk_scripts.js",
]


def setup(app):
    app.add_css_file("custom.css")


# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- Options for breathe --------------------------------------------------

# Enable the breathe extension
extensions.append("breathe")
extensions.append("exhale")

# Set up the default project for breathe extension
breathe_default_project = "cvcuda"


# -- Options for sphinx_rtd_theme -----------------------------------------

# Enable the sphinx_rtd_theme extension
extensions.append("sphinx_rtd_theme")

# Enable the sphinx.ext.todo extension
extensions.append("sphinx.ext.todo")

# -- Extension configuration -------------------------------------------------

doxygen_input_config = """
INPUT                = ../../src/core/include
INPUT               += ../../src/operators/include
INPUT               += ../../src/optools/include
INPUT               += ../../src/format/include
"""
doxygen_config = (
    doxygen_input_config
    + """
EXCLUDE             += ../../tests
EXCLUDE_PATTERNS     = *.md *.txt
ENABLE_PREPROCESSING = YES
WARN_IF_UNDOCUMENTED = NO
USE_M
"""
)

doxygen_conf_extra = """
INLINE_SIMPLE_STRUCTS = YES
TYPEDEF_HIDES_STRUCT = YES
EXPAND_ONLY_PREDEF = YES
"""

doxygen_predefined = [
    "NVCV_PUBLIC=",
    "NVCV_API_VERSION_IS(x,y)=0",
    "NVCV_API_VERSION_AT_LEAST(x,y)=1",
    "NVCV_API_VERSION_AT_MOST(x,y)=0",
]

# Setup the exhale extension
exhale_args = {
    # These arguments are required
    "containmentFolder": "_exhale_api",
    "rootFileName": "cvcuda_api.rst",
    "doxygenStripFromPath": "../../src",
    # Heavily encouraged optional argument (see docs)
    "rootFileTitle": "Library API",
    # Suggested optional arguments
    "createTreeView": True,
    # TIP: if using the sphinx-bootstrap-theme, you need
    # "treeViewIsBootstrap": True,
    "exhaleExecutesDoxygen": False,
    "fullToctreeMaxDepth": 1,
    "minifyTreeView": False,
    "contentsDirectives": False,
    "exhaleDoxygenStdin": doxygen_input_config,
}

# Tell sphinx what the primary language being documented is.
primary_domain = "cpp"

# Tell sphinx what the pygments highlight language should be.
highlight_language = "cpp"



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cvcuda
import pytest as t
import numpy as np
import util


RNG = np.random.default_rng(0)


@t.mark.parametrize(
    "input, ksize",
    [
        (
            cvcuda.Tensor([5, 9, 9, 4], np.uint8, "NHWC"),
            [5, 5],
        ),
        (
            cvcuda.Tensor([9, 9, 3], np.uint8, "HWC"),
            [5, 5],
        ),
        (
            cvcuda.Tensor([5, 21, 21, 4], np.uint8, "NHWC"),
            [15, 15],
        ),
        (
            cvcuda.Tensor([21, 21, 3], np.uint8, "HWC"),
            [15, 15],
        ),
    ],
)
def test_op_median_blur(input, ksize):
    out = cvcuda.median_blur(input, ksize)
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == input.dtype

    stream = cvcuda.Stream()
    out = cvcuda.Tensor(input.shape, input.dtype, input.layout)
    tmp = cvcuda.median_blur_into(
        src=input,
        dst=out,
        ksize=ksize,
        stream=stream,
    )
    assert tmp is out
    assert out.layout == input.layout
    assert out.shape == input.shape
    assert out.dtype == input.dtype


@t.mark.parametrize(
    "nimages, format, max_size, max_pixel, max_ksize",
    [
        (
            5,
            cvcuda.Format.RGB8,
            (16, 23),
            128.0,
            [11, 11],
        ),
        (
            5,
            cvcuda.Format.RGB8,
            (16, 23),
            128.0,
            [25, 25],
        ),
    ],
)
def test_op_median_blurvarshape(nimages, format, max_size, max_pixel, max_ksize):

    input = util.create_image_batch(
        nimages, format, max_size=max_size, max_random=max_pixel, rng=RNG
    )

    ksize = util.create_tensor(
        (nimages, 2),
        np.int32,
        "NC",
        max_random=max_ksize,
        rng=RNG,
        transform_dist=util.dist_odd,
    )

    out = cvcuda.median_blur(input, ksize)
    assert len(out) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == input.uniqueformat
    assert out.maxsize == input.maxsize

    stream = cvcuda.Stream()

    out = util.clone_image_batch(input)
    tmp = cvcuda.median_blur_into(
        src=input,
        dst=out,
        ksize=ksize,
        stream=stream,
    )
    assert tmp is out
    assert len(out) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == input.uniqueformat
    assert out.maxsize == input.maxsize



# SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cvcuda
import pytest as t
import numpy as np
import util


RNG = np.random.default_rng(0)


@t.mark.parametrize(
    "input, kernel, kernel_anchor, border",
    [
        (
            util.create_image_batch(
                10, cvcuda.Format.RGB8, size=(123, 321), max_random=256, rng=RNG
            ),
            util.create_image_batch(
                10, cvcuda.Format.F32, size=(3, 3), max_random=1, rng=RNG
            ),
            util.create_tensor((10, 2), np.int32, "NC", max_random=(3, 3), rng=RNG),
            cvcuda.Border.CONSTANT,
        ),
        (
            util.create_image_batch(7, cvcuda.Format.RGBf32, max_random=1, rng=RNG),
            util.create_image_batch(
                7, cvcuda.Format.F32, size=(5, 5), max_random=3, rng=RNG
            ),
            util.create_tensor((7, 2), np.int32, "NC", max_random=(5, 5), rng=RNG),
            cvcuda.Border.REPLICATE,
        ),
        (
            util.create_image_batch(1, cvcuda.Format.U8, max_random=123, rng=RNG),
            util.create_image_batch(
                1, cvcuda.Format.F32, size=(7, 7), max_random=2, rng=RNG
            ),
            util.create_tensor((1, 2), np.int32, "NC", max_random=(7, 7), rng=RNG),
            cvcuda.Border.REFLECT,
        ),
        (
            util.create_image_batch(6, cvcuda.Format.S16, max_random=1234, rng=RNG),
            util.create_image_batch(
                6, cvcuda.Format.F32, max_size=(9, 9), max_random=4, rng=RNG
            ),
            util.create_tensor((6, 2), np.int32, "NC", max_random=(1, 1), rng=RNG),
            cvcuda.Border.WRAP,
        ),
        (
            util.create_image_batch(9, cvcuda.Format.S32, max_random=12345, rng=RNG),
            util.create_image_batch(
                9, cvcuda.Format.F32, max_size=(4, 4), max_random=2, rng=RNG
            ),
            util.create_tensor((9, 2), np.int32, "NC", max_random=(4, 4), rng=RNG),
            cvcuda.Border.REFLECT101,
        ),
    ],
)
def test_op_conv2dvarshape(input, kernel, kernel_anchor, border):
    out = cvcuda.conv2d(
        input,
        kernel,
        kernel_anchor,
        border,
    )
    assert len(out) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == input.uniqueformat
    assert out.maxsize == input.maxsize

    stream = cvcuda.Stream()
    out = util.clone_image_batch(input)
    tmp = cvcuda.conv2d_into(
        src=input,
        dst=out,
        kernel=kernel,
        kernel_anchor=kernel_anchor,
        border=border,
        stream=stream,
    )
    assert tmp is out
    assert len(out) == len(input)
    assert out.capacity == input.capacity
    assert out.uniqueformat == input.uniqueformat
    assert out.maxsize == input.maxsize
